/* life.h: Generic application-wide definitions.			*/
/*									*/
/* (C) 1997 by Brian Raiter, under the terms of the GNU General Public	*/
/* License ver. 2 (or any later version, if you like).			*/

#ifndef __LIFE_H__
#define __LIFE_H__


#include	"lifeids.h"

/* Generic out-of-bounds value. */
#define	NIL		-1

/* Commands sent to the various Set functions. */
#define	SET_INIT	1
#define	SET_ALTER	2
#define	SET_END		3

/* Upper and lower bounds on the world sizes. */
#define	MIN_WORLDSIDE	1
#define	MIN_WORLDSIZE	(1 * 1)
#define	MAX_WORLDSIDE	255
#define	MAX_WORLDSIZE	(255 * 255)

/* Flags sent to the file I/O functions. */
#define	WORLD_IN	0x8000
#define	WORLD_INTO	0x8001
#define	WORLD_INMERGE	0x8002
#define	WORLD_OUT	0x4000
#define	WORLD_NOPROMPT	0x0010

/* Color indexes for the various window elements. */
#define	EL_BKGND	0
#define	EL_RECT		1
#define	EL_TEXT		2
#define	EL_DEAD		3
#define	EL_ALIVE	4
#define	EL_ZYGOTE	5
#define	EL_DYING	6
#define	NUM_ELS		7
#define	EL_COLLAPSIBLE	EL_TEXT

/* The different file types. */
#define	EXT_BMP		0
#define	EXT_LIFETEXT	1
#define	EXT_TEXT	2
#define	EXT_UNKNOWN	3

/* Macros to translate the data stored in a cell. */
#define	CellAlive(C)		((BOOL)((C) & 0x80))
#define	CellNeighbors(C)	((int)(((C) & 0x7F) % 9))
#define	CellDeferred(C)		((int)(((C) & 0x7F) / 9) - 4)
#define	Live(C)			((C) |= 0x80)
#define	Die(C)			((C) &= 0x7F)
#define	MakeCell(A, N, D)	((BYTE)(((A) ? 0x80 : 0) +		\
					(((N) + ((D) + 4) * 9) & 0x7F)))

/* Generic utility macros. */
#define	minmax(lo, x, hi)	((x) < (lo) ? (lo) : (x) > (hi) ? (hi) : (x))
#define	GetString(id)							\
		(LoadString(hinstApp, (id), szBuffer, sizeof szBuffer))
#define	MakeDlgProcInstance(lpfn, hinst)				\
		((DLGPROC)MakeProcInstance((FARPROC)(lpfn), (hinst)))
#define	FreeDlgProcInstance(lpfn)					\
		(FreeProcInstance((FARPROC)(lpfn)))
#define HANDLE_DLGMSG(hwnd, message, fn)				\
    case (message): return (BOOL)HANDLE_##message((hwnd),		\
					(wParam), (lParam), (fn))

/* User-defined messages. */
#define	WM_SYSTEMIDLE	(WM_USER + WM_ENTERIDLE)
#define	HANDLE_WM_SYSTEMIDLE(hwnd, wParam, lParam, fn)			\
    ((fn)(hwnd), 0L)
#define	FORWARD_WM_SYSTEMIDLE(hwnd, fn)					\
    (void)(fn)((hwnd), WM_SYSTEMIDLE, 0, 0L)
#define	WM_EXITTIME	(WM_USER + WM_TIMER)
#define	HANDLE_WM_EXITTIME(hwnd, wParam, lParam, fn)			\
    ((fn)(hwnd), 0L)
#define	FORWARD_WM_EXITTIME(hwnd, fn)					\
    (void)(fn)((hwnd), WM_EXITTIME, 0, 0L)
#define	WM_DELETEOBJECT	(WM_USER + WM_DESTROY)
#define	HANDLE_WM_DELETEOBJECT(hwnd, wParam, lParam, fn)		\
    ((fn)((hwnd), (HGDIOBJ)(wParam)), 0L)
#define	FORWARD_WM_DELETEOBJECT(hwnd, hgdiobj, fn)			\
    (void)(fn)((hwnd), WM_DELETEOBJECT, (WPARAM)(hgdiobj), 0L)

/* Generic typedefs. */
typedef	void NEAR* PVOID;
typedef	void _huge* HPVOID;
typedef void const FAR* LPCVOID;
typedef	void const NEAR* PCVOID;
typedef void const _huge* HPCVOID;
typedef	LPVOID FAR* LPLPVOID;
typedef	UINT FAR* LPUINT;
typedef DWORD const FAR* LPCDWORD;
typedef COLORREF FAR* LPCOLORREF;

/* What the world looks like. */
typedef	BYTE (* PWORLD)[256];
typedef	BYTE (FAR* LPWORLD)[256];

/* A 8-bit BITMAPINFO structure that gives the right sizeof. */
typedef	struct {
		BITMAPINFOHEADER	bmih;
		int			idx[256];
	} DIB256INFO, FAR* LPDIB256INFO;

/* A union of all the different color representations. */
typedef	union {
		COLORREF			ref;
		PALETTEENTRY			pal;
		struct { BYTE r, g, b, f; }	rgb;
		struct { BYTE h, l, s, f; }	hls;
		BYTE				val[4];
	} COLOR, * PCOLOR, FAR* LPCOLOR;

/* Special "friend" function, allowing lifecalc to call into lifedisp. */
extern void SetDisplaySpace(int, int, UINT);


#endif	/*  __LIFE_H__  */
